Instructions to use replicate/flashinfer-draft with libraries, inference providers, notebooks, and local apps. Follow these links to get started.
- Libraries
- Kernels
How to use replicate/flashinfer-draft with Kernels:
# !pip install kernels from kernels import get_kernel # a version (or an explicit revision) is required; see the "Files and versions" tab for the available ones kernel = get_kernel("replicate/flashinfer-draft", version=1) - Notebooks
- Google Colab
- Kaggle
File size: 34,254 Bytes
57c3a10 | 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 264 265 266 267 268 269 270 271 272 273 274 275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 302 303 304 305 306 307 308 309 310 311 312 313 314 315 316 317 318 319 320 321 322 323 324 325 326 327 328 329 330 331 332 333 334 335 336 337 338 339 340 341 342 343 344 345 346 347 348 349 350 351 352 353 354 355 356 357 358 359 360 361 362 363 364 365 366 367 368 369 370 371 372 373 374 375 376 377 378 379 380 381 382 383 384 385 386 387 388 389 390 391 392 393 394 395 396 397 398 399 400 401 402 403 404 405 406 407 408 409 410 411 412 413 414 415 416 417 418 419 420 421 422 423 424 425 426 427 428 429 430 431 432 433 434 435 436 437 438 439 440 441 442 443 444 445 446 447 448 449 450 451 452 453 454 455 456 457 458 459 460 461 462 463 464 465 466 467 468 469 470 471 472 473 474 475 476 477 478 479 480 481 482 483 484 485 486 487 488 489 490 491 492 493 494 495 496 497 498 499 500 501 502 503 504 505 506 507 508 509 510 511 512 513 514 515 516 517 518 519 520 521 522 523 524 525 526 527 528 529 530 531 532 533 534 535 536 537 538 539 540 541 542 543 544 545 546 547 548 549 550 551 552 553 554 555 556 557 558 559 560 561 562 563 564 565 566 567 568 569 570 571 572 573 574 575 576 577 578 579 580 581 582 583 584 585 586 587 588 589 590 591 592 593 594 595 596 597 598 599 600 601 602 603 604 605 606 607 608 609 610 611 612 613 614 615 616 617 618 619 620 621 622 623 624 625 626 627 628 629 630 631 632 633 634 635 636 637 638 639 640 641 642 643 644 645 646 647 648 649 650 651 652 653 654 655 656 657 658 659 660 661 662 663 664 665 666 667 668 669 670 671 672 673 674 675 676 677 678 679 680 681 682 683 684 685 686 687 688 689 690 691 692 693 694 695 696 697 698 699 700 701 702 703 704 705 706 707 708 709 710 711 712 713 714 715 716 717 718 719 720 721 722 723 724 725 726 727 728 729 730 731 732 733 734 735 736 737 738 739 740 741 742 743 744 745 746 747 748 749 750 751 752 753 754 755 756 757 758 759 760 761 762 763 764 765 766 767 768 769 770 771 772 773 774 775 776 777 778 779 780 781 782 783 784 785 786 787 | /*!
* Copyright (c) 2023 by FlashInfer team.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
#ifndef FLASHINFER_CASCADE_CUH_
#define FLASHINFER_CASCADE_CUH_
#include "../cp_async.cuh"
#include "../math.cuh"
#include "../utils.cuh"
#include "state.cuh"
namespace flashinfer {
using cp_async::PrefetchMode;
using cp_async::SharedMemFillMode;
/*!
* \brief The CUDA kernel that merges the self-attention state of two index sets A and B.
* \tparam vec_size The vector size used in the kernel.
* \tparam DTypeIn The data type of v_a and v_b.
* \tparam DTypeO The data type of v_merged.
* \param v_a The partial v of index set A. (n, h, d)
* \param s_a The logsumexp value of index set A. (n, h)
* \param v_b The partial v of index set B. (n, h, d)
* \param s_b The logsumexp value of index set B. (n, h)
* \param v_merged The merged v of index set A union B. (n, h, d)
* \param s_merged The merged logsumexp value of index set A union B. (n, h)
* \param num_heads The number of heads of v_a and v_b.
* \param head_dim The dimension of each head.
* \note Both s_a and s_b are logsumexp values with base 2.
*/
template <uint32_t vec_size, typename DTypeIn, typename DTypeO>
__global__ void MergeStateKernel(DTypeIn* __restrict__ v_a, float* __restrict__ s_a,
DTypeIn* __restrict__ v_b, float* __restrict__ s_b,
DTypeO* __restrict__ v_merged, float* __restrict__ s_merged,
uint32_t num_heads, uint32_t head_dim) {
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t pos = blockIdx.x;
uint32_t head_idx = ty;
float s_a_val = s_a[pos * num_heads + head_idx];
float s_b_val = s_b[pos * num_heads + head_idx];
float s_max = max(s_a_val, s_b_val);
s_a_val = math::ptx_exp2(s_a_val - s_max);
s_b_val = math::ptx_exp2(s_b_val - s_max);
float a_scale = s_a_val / (s_a_val + s_b_val);
float b_scale = s_b_val / (s_a_val + s_b_val);
vec_t<float, vec_size> v_a_vec, v_b_vec, v_merged_vec;
v_a_vec.cast_load(v_a + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
v_b_vec.cast_load(v_b + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
#pragma unroll
for (uint32_t i = 0; i < vec_size; ++i) {
v_merged_vec[i] = a_scale * v_a_vec[i] + b_scale * v_b_vec[i];
}
v_merged_vec.cast_store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = math::ptx_log2(s_a_val + s_b_val) + s_max;
}
}
/*!
* \brief The CUDA kernel that merges the self-attention state with another state in-place.
* \tparam vec_size The vector size used in the kernel.
* \tparam DType The data type of v and v_other.
* \param v The partial v to be updated in-place. (n, h, d)
* \param s The logsumexp value to be updated in-place. (n, h)
* \param v_other The other v to be merged. (n, h, d)
* \param s_other The other logsumexp value to be merged. (n, h)
* \param mask Optional mask of whether to merge given sequences or not. (n)
* \param num_heads The number of heads of v and v_other.
* \param head_dim The dimension of each head.
* \note Both s and s_other are logsumexp values with base 2.
*/
template <uint32_t vec_size, typename DType>
__global__ void MergeStateInPlaceKernel(DType* __restrict__ v, float* __restrict__ s,
DType* __restrict__ v_other, float* __restrict__ s_other,
uint8_t* __restrict__ mask, uint32_t num_heads,
uint32_t head_dim) {
uint32_t pos = blockIdx.x;
if (mask != nullptr && mask[pos] == 0) return;
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t head_idx = ty;
float s_val = s[pos * num_heads + head_idx];
float s_other_val = s_other[pos * num_heads + head_idx];
float s_max = max(s_val, s_other_val);
s_val = math::ptx_exp2(s_val - s_max);
s_other_val = math::ptx_exp2(s_other_val - s_max);
float scale = s_val / (s_val + s_other_val);
float other_scale = s_other_val / (s_val + s_other_val);
vec_t<float, vec_size> v_vec, v_other_vec;
v_vec.cast_load(v + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
v_other_vec.cast_load(v_other + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
#pragma unroll
for (uint32_t i = 0; i < vec_size; ++i) {
v_vec[i] = scale * v_vec[i] + other_scale * v_other_vec[i];
}
v_vec.cast_store(v + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s != nullptr) {
s[pos * num_heads + head_idx] = math::ptx_log2(s_val + s_other_val) + s_max;
}
}
template <uint32_t bdx, uint32_t bdy, uint32_t vec_size, typename DTypeIn>
__device__ __forceinline__ void threadblock_sync_state(state_t<vec_size>& st, DTypeIn* v_smem,
float* s_smem,
const uint32_t tx = threadIdx.x,
const uint32_t ty = threadIdx.y) {
constexpr uint32_t head_dim = vec_size * bdx;
st.o.cast_store(v_smem + ty * head_dim + tx * vec_size);
s_smem[ty] = st.get_lse();
st.init();
__syncthreads();
#pragma unroll
for (uint32_t iter = 0; iter < bdy; ++iter) {
float s = s_smem[iter];
vec_t<float, vec_size> v;
v.cast_load(v_smem + iter * head_dim + tx * vec_size);
st.merge(v, s, 1);
}
}
template <uint32_t bdx, uint32_t bdy, uint32_t vec_size, typename DTypeIn>
__device__ __forceinline__ void warp_sync_state(state_t<vec_size>& st, DTypeIn* v_smem,
float* s_smem, const uint32_t tx = threadIdx.x,
const uint32_t ty = threadIdx.y) {
constexpr uint32_t head_dim = vec_size * bdx;
st.o.cast_store(v_smem + ty * head_dim + tx * vec_size);
s_smem[ty] = st.get_lse();
st.init();
__syncwarp();
#pragma unroll
for (uint32_t iter = 0; iter < bdy; ++iter) {
float s = s_smem[iter];
vec_t<float, vec_size> v;
v.cast_load(v_smem + iter * head_dim + tx * vec_size);
st.merge(v, s, 1);
}
}
template <uint32_t bdx, uint32_t bdy, uint32_t vec_size, typename DTypeIn>
__device__ __forceinline__ void threadblock_sum(vec_t<float, vec_size>& v, DTypeIn* v_smem) {
const uint32_t tx = threadIdx.x, ty = threadIdx.y;
constexpr uint32_t head_dim = vec_size * bdx;
v.cast_store(v_smem + ty * head_dim + tx * vec_size);
v.fill(DTypeIn(0.f));
__syncthreads();
#pragma unroll
for (uint32_t iter = 0; iter < bdy; ++iter) {
vec_t<float, vec_size> v_iter;
v_iter.cast_load(v_smem + iter * head_dim + tx * vec_size);
#pragma unroll
for (uint32_t i = 0; i < vec_size; ++i) {
v[i] += v_iter[i];
}
}
}
template <uint32_t vec_size, typename DTypeIn, typename DTypeO>
__global__ void AttentionSumKernel(DTypeIn* __restrict__ V, DTypeO* __restrict__ v_sum,
uint32_t num_index_sets, uint32_t num_heads, uint32_t head_dim) {
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t pos = blockIdx.x;
uint32_t head_idx = ty;
if (num_index_sets == 0) {
vec_t<DTypeO, vec_size> v;
v.fill(DTypeO(0.f));
v.store(v_sum + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
return;
}
if (num_index_sets == 1) {
vec_t<DTypeO, vec_size> v;
v.cast_load(V + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
v.store(v_sum + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
return;
}
vec_t<float, vec_size> v_sum_vec;
v_sum_vec.fill(0.f);
#pragma unroll 2
for (uint32_t iter = 0; iter < num_index_sets; ++iter) {
vec_t<float, vec_size> v;
v.cast_load(V + ((pos * num_index_sets + iter) * num_heads + head_idx) * head_dim +
tx * vec_size);
#pragma unroll
for (uint32_t i = 0; i < vec_size; ++i) {
v_sum_vec[i] += v[i];
}
}
v_sum_vec.cast_store(v_sum + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
}
template <uint32_t vec_size, typename DTypeIn, typename DTypeO>
__global__ void MergeStatesKernel(DTypeIn* __restrict__ V, float* __restrict__ S,
DTypeO* __restrict__ v_merged, float* __restrict__ s_merged,
uint32_t num_index_sets, uint32_t num_heads, uint32_t head_dim) {
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t pos = blockIdx.x;
uint32_t head_idx = ty;
if (num_index_sets == 0) {
vec_t<DTypeO, vec_size> v;
v.fill(DTypeO(0.f));
v.store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = -math::inf;
}
return;
}
if (num_index_sets == 1) {
vec_t<DTypeO, vec_size> v;
v.cast_load(V + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
v.store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = S[pos * num_heads + head_idx];
}
return;
}
state_t<vec_size> st;
#pragma unroll 2
for (uint32_t iter = 0; iter < num_index_sets; ++iter) {
float s = S[(pos * num_index_sets + iter) * num_heads + head_idx];
vec_t<float, vec_size> v;
v.cast_load(V + ((pos * num_index_sets + iter) * num_heads + head_idx) * head_dim +
tx * vec_size);
st.merge(v, s, 1);
}
st.normalize();
st.o.cast_store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = st.get_lse();
}
}
/*!
* \brief The CUDA kernel that merges self-attention states of a list of index sets,
* accelerated for larger number of index sets.
* \tparam vec_size The vector size used in the kernel.
* \tparam bdx The blockDim.x used in the kernel.
* \tparam bdy The blockDim.y used in the kernel.
* \tparam num_smem_stages The number of stages of shared memory used in the kernel.
* \tparam DTypeIn The data type of v.
* \tparam DTypeO The data type of v_merged.
* \param V The partial v of index sets. (n, num_index_sets, h, d)
* \param S The logsumexp value of index sets. (n, num_index_sets, h)
* \param v_merged The merged v of index sets union. (n, h, d)
* \param s_merged The merged logsumexp value of index sets union. (n, h)
* \param num_heads The number of heads of v.
* \param head_dim The dimension of each head.
* \note s are logsumexp values with base 2.
*/
template <uint32_t vec_size, uint32_t bdx, uint32_t bdy, uint32_t num_smem_stages, typename DTypeIn,
typename DTypeO>
__global__ void MergeStatesLargeNumIndexSetsKernel(DTypeIn* __restrict__ V, float* __restrict__ S,
DTypeO* __restrict__ v_merged,
float* __restrict__ s_merged,
uint32_t num_index_sets, uint32_t num_heads) {
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t pos = blockIdx.x;
uint32_t head_idx = blockIdx.y;
state_t<vec_size> st;
constexpr uint32_t vec_bits = sizeof(DTypeIn) * vec_size * 8;
constexpr uint32_t head_dim = vec_size * bdx;
extern __shared__ uint8_t smem[];
DTypeIn* v_smem = (DTypeIn*)smem;
float* s_smem = (float*)(smem + num_smem_stages * bdy * head_dim * sizeof(DTypeIn));
#pragma unroll
for (uint32_t iter = 0; iter < num_smem_stages; ++iter) {
cp_async::pred_load<vec_bits, PrefetchMode::kPrefetch, SharedMemFillMode::kNoFill>(
v_smem + (iter * bdy + ty) * head_dim + tx * vec_size,
V + ((pos * num_index_sets + (iter * bdy + ty)) * num_heads + head_idx) * head_dim +
tx * vec_size,
(iter * bdy + ty) < num_index_sets);
cp_async::commit_group();
}
#pragma unroll 4
for (uint32_t iter = 0; iter < ceil_div(num_index_sets, bdy); ++iter) {
if (iter % bdx == 0) {
s_smem[ty * bdx + tx] =
iter * bdy + (ty * bdx + tx) < num_index_sets
? S[(pos * num_index_sets + (iter * bdy + ty * bdx + tx)) * num_heads + head_idx]
: 0.f;
__syncthreads();
}
cp_async::wait_group<num_smem_stages - 1>();
__syncthreads();
vec_t<float, vec_size> v;
v.cast_load(v_smem + ((iter % num_smem_stages) * bdy + ty) * head_dim + tx * vec_size);
if (iter * bdy + ty < num_index_sets) {
float s = s_smem[(iter % bdx) * bdy + ty];
st.merge(v, s, 1);
}
__syncthreads();
cp_async::pred_load<vec_bits, PrefetchMode::kPrefetch, SharedMemFillMode::kNoFill>(
v_smem + ((iter % num_smem_stages) * bdy + ty) * head_dim + tx * vec_size,
V +
((pos * num_index_sets + ((iter + num_smem_stages) * bdy + ty)) * num_heads +
head_idx) *
head_dim +
tx * vec_size,
(iter + num_smem_stages) * bdy + ty < num_index_sets);
cp_async::commit_group();
}
cp_async::wait_group<0>();
__syncthreads();
st.normalize();
threadblock_sync_state<bdx, bdy, vec_size>(st, v_smem, s_smem);
st.normalize();
st.o.cast_store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = st.get_lse();
}
}
/*!
* \brief The CUDA kernel to merge self-attention states of multiple index sets, the number of
* index sets at each position might vary.
*
* For CUDA graph support, the kernel can be built with a maximum sequence length and executed
* using a truncated, dynamic sequence length passed through `seq_len_ptr`.
*
* \tparam vec_size The vector size used in the kernel.
* \tparam bdx The blockDim.x used in the kernel.
* \tparam bdy The blockDim.y used in the kernel.
* \tparam num_smem_stages The number of stages of shared memory used in the kernel.
* \tparam DTypeIn The data type of v.
* \tparam DTypeO The data type of v_merged.
* \param V The partial v of index sets. (nnz, h, d)
* \param S The logsumexp value of index sets. (nnz, h)
* \param indptr The start offsets of each position in the variable length array.
* \param v_merged The merged v of index sets union. (n, h, d)
* \param s_merged The merged logsumexp value of index sets union. (n, h)
* \param max_seq_len The maximum sequence length supported by the kernel.
* \param seq_len_ptr The current sequence length (number of positions populated in indptr).
* \param num_heads The number of heads of v.
* \param head_dim The dimension of each head.
* \note s are logsumexp values with base 2.
*/
template <uint32_t vec_size, uint32_t bdx, uint32_t bdy, uint32_t num_smem_stages, typename DTypeIn,
typename DTypeO, typename IdType>
__global__ void PersistentVariableLengthMergeStatesKernel(
DTypeIn* __restrict__ V, float* __restrict__ S, IdType* indptr, DTypeO* __restrict__ v_merged,
float* __restrict__ s_merged, uint32_t max_seq_len, uint32_t* __restrict__ seq_len_ptr,
uint32_t num_heads) {
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t cta_id = blockIdx.x;
uint32_t num_ctas = gridDim.x;
const uint32_t seq_len = seq_len_ptr ? *seq_len_ptr : max_seq_len;
uint32_t num_iters = ceil_div(seq_len * num_heads, num_ctas);
constexpr uint32_t vec_bits = sizeof(DTypeIn) * vec_size * 8;
constexpr uint32_t head_dim = vec_size * bdx;
extern __shared__ uint8_t smem[];
DTypeIn* v_smem = (DTypeIn*)smem;
float* s_smem = (float*)(smem + num_smem_stages * bdy * head_dim * sizeof(DTypeIn));
#if (__CUDACC_VER_MAJOR__ >= 12 && defined(__CUDA_ARCH__) && (__CUDA_ARCH__ >= 900))
asm volatile("griddepcontrol.wait;");
#endif
#pragma unroll 1
for (uint32_t i = cta_id; i < seq_len * num_heads; i += num_ctas) {
uint32_t pos = i / num_heads;
uint32_t head_idx = i % num_heads;
state_t<vec_size> st;
const uint32_t num_index_sets = indptr[pos + 1] - indptr[pos];
if (num_index_sets == 0) {
vec_t<DTypeO, vec_size> v;
v.fill(DTypeO(0.f));
v.store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = -math::inf;
}
continue;
}
if (num_index_sets == 1) {
vec_t<DTypeO, vec_size> v;
v.cast_load(V + (indptr[pos] * num_heads + head_idx) * head_dim + tx * vec_size);
v.store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = S[indptr[pos] * num_heads + head_idx];
}
continue;
}
#pragma unroll
for (uint32_t iter = 0; iter < num_smem_stages; ++iter) {
cp_async::pred_load<vec_bits, PrefetchMode::kPrefetch, SharedMemFillMode::kNoFill>(
v_smem + (iter * bdy + ty) * head_dim + tx * vec_size,
V + ((indptr[pos] + (iter * bdy + ty)) * num_heads + head_idx) * head_dim + tx * vec_size,
(iter * bdy + ty) < num_index_sets);
cp_async::commit_group();
}
#pragma unroll 4
for (uint32_t iter = 0; iter < ceil_div(num_index_sets, bdy); ++iter) {
if (iter % bdx == 0) {
s_smem[ty * bdx + tx] =
iter * bdy + (ty * bdx + tx) < num_index_sets
? S[(indptr[pos] + (iter * bdy + ty * bdx + tx)) * num_heads + head_idx]
: 0.f;
__syncthreads();
}
cp_async::wait_group<num_smem_stages - 1>();
__syncthreads();
vec_t<float, vec_size> v;
v.cast_load(v_smem + ((iter % num_smem_stages) * bdy + ty) * head_dim + tx * vec_size);
if (iter * bdy + ty < num_index_sets) {
float s = s_smem[(iter % bdx) * bdy + ty];
st.merge(v, s, 1);
}
__syncthreads();
cp_async::pred_load<vec_bits, PrefetchMode::kPrefetch, SharedMemFillMode::kNoFill>(
v_smem + ((iter % num_smem_stages) * bdy + ty) * head_dim + tx * vec_size,
V +
((indptr[pos] + ((iter + num_smem_stages) * bdy + ty)) * num_heads + head_idx) *
head_dim +
tx * vec_size,
(iter + num_smem_stages) * bdy + ty < num_index_sets);
cp_async::commit_group();
}
cp_async::wait_group<0>();
__syncthreads();
st.normalize();
threadblock_sync_state<bdx, bdy, vec_size>(st, v_smem, s_smem);
st.normalize();
st.o.cast_store(v_merged + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
if (s_merged != nullptr) {
s_merged[pos * num_heads + head_idx] = st.get_lse();
}
}
#if (__CUDACC_VER_MAJOR__ >= 12 && defined(__CUDA_ARCH__) && (__CUDA_ARCH__ >= 900))
asm volatile("griddepcontrol.launch_dependents;");
#endif
}
template <uint32_t vec_size, uint32_t bdx, uint32_t bdy, uint32_t num_smem_stages, typename DTypeIn,
typename DTypeO, typename IdType>
__global__ void PersistentVariableLengthAttentionSumKernel(DTypeIn* __restrict__ V, IdType* indptr,
DTypeO* __restrict__ v_sum,
uint32_t max_seq_len,
uint32_t* __restrict__ seq_len_ptr,
uint32_t num_heads) {
uint32_t tx = threadIdx.x, ty = threadIdx.y;
uint32_t cta_id = blockIdx.x;
uint32_t num_ctas = gridDim.x;
const uint32_t seq_len = seq_len_ptr ? *seq_len_ptr : max_seq_len;
uint32_t num_iters = ceil_div(seq_len * num_heads, num_ctas);
constexpr uint32_t vec_bits = sizeof(DTypeIn) * vec_size * 8;
constexpr uint32_t head_dim = vec_size * bdx;
extern __shared__ uint8_t smem[];
DTypeIn* v_smem = (DTypeIn*)smem;
vec_t<float, vec_size> v_sum_vec;
#if (__CUDACC_VER_MAJOR__ >= 12 && defined(__CUDA_ARCH__) && (__CUDA_ARCH__ >= 900))
asm volatile("griddepcontrol.wait;");
#endif
#pragma unroll 1
for (uint32_t i = cta_id; i < seq_len * num_heads; i += num_ctas) {
uint32_t pos = i / num_heads;
uint32_t head_idx = i % num_heads;
const uint32_t num_index_sets = indptr[pos + 1] - indptr[pos];
if (num_index_sets == 0) {
vec_t<DTypeO, vec_size> v;
v.fill(DTypeO(0.f));
v.store(v_sum + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
continue;
}
if (num_index_sets == 1) {
vec_t<DTypeO, vec_size> v;
v.cast_load(V + (indptr[pos] * num_heads + head_idx) * head_dim + tx * vec_size);
v.store(v_sum + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
continue;
}
#pragma unroll
for (uint32_t iter = 0; iter < num_smem_stages; ++iter) {
cp_async::pred_load<vec_bits, PrefetchMode::kPrefetch, SharedMemFillMode::kNoFill>(
v_smem + (iter * bdy + ty) * head_dim + tx * vec_size,
V + ((indptr[pos] + (iter * bdy + ty)) * num_heads + head_idx) * head_dim + tx * vec_size,
(iter * bdy + ty) < num_index_sets);
cp_async::commit_group();
}
#pragma unroll 4
for (uint32_t iter = 0; iter < ceil_div(num_index_sets, bdy); ++iter) {
cp_async::wait_group<num_smem_stages - 1>();
__syncthreads();
vec_t<float, vec_size> v;
v.cast_load(v_smem + ((iter % num_smem_stages) * bdy + ty) * head_dim + tx * vec_size);
if (iter * bdy + ty < num_index_sets) {
#pragma unroll
for (uint32_t i = 0; i < vec_size; ++i) {
v_sum_vec[i] += v[i];
}
}
__syncthreads();
cp_async::pred_load<vec_bits, PrefetchMode::kPrefetch, SharedMemFillMode::kNoFill>(
v_smem + ((iter % num_smem_stages) * bdy + ty) * head_dim + tx * vec_size,
V +
((indptr[pos] + ((iter + num_smem_stages) * bdy + ty)) * num_heads + head_idx) *
head_dim +
tx * vec_size,
(iter + num_smem_stages) * bdy + ty < num_index_sets);
cp_async::commit_group();
}
cp_async::wait_group<0>();
__syncthreads();
threadblock_sum<bdx, bdy, vec_size>(v_sum_vec, v_smem);
v_sum_vec.cast_store(v_sum + (pos * num_heads + head_idx) * head_dim + tx * vec_size);
}
#if (__CUDACC_VER_MAJOR__ >= 12 && defined(__CUDA_ARCH__) && (__CUDA_ARCH__ >= 900))
asm volatile("griddepcontrol.launch_dependents;");
#endif
}
/*!
* \brief Merge the self-attention state of two index sets A and B.
* \tparam DTypeIn The data type of v_a and v_b.
* \tparam DTypeO The data type of v_merged.
* \param v_a The partial v of index set A (n, h, d)
* \param s_a The logsumexp value of index set A. (n, h)
* \param v_b The partial v of index set B. (n, h, d)
* \param s_b The logsumexp value of index set B. (n, h)
* \param v_merged The merged v of index set A union B. (n, h, d)
* \param s_merged The merged logsumexp value of index set A union B. (n, h)
* \param seq_len The sequence length.
* \param num_heads The number of heads of v_a and v_b.
* \param head_dim The dimension of each head.
* \param stream The CUDA stream to execute the kernel.
* \return status Indicates whether CUDA calls are successful
* \note Both s_a and s_b are logsumexp values with base 2.
*/
template <typename DTypeIn, typename DTypeO>
cudaError_t MergeState(DTypeIn* v_a, float* s_a, DTypeIn* v_b, float* s_b, DTypeO* v_merged,
float* s_merged, uint32_t seq_len, uint32_t num_heads, uint32_t head_dim,
cudaStream_t stream = nullptr) {
DISPATCH_HEAD_DIM(head_dim, HEAD_DIM, {
constexpr uint32_t vec_size = std::max(16U / sizeof(DTypeIn), HEAD_DIM / 32U);
uint32_t bdx = HEAD_DIM / vec_size;
uint32_t bdy = num_heads;
dim3 nblks(seq_len);
dim3 nthrs(bdx, bdy);
auto kernel = MergeStateKernel<vec_size, DTypeIn, DTypeO>;
void* args[] = {&v_a, &s_a, &v_b, &s_b, &v_merged, &s_merged, &num_heads, &head_dim};
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, 0, stream));
});
return cudaSuccess;
}
/*!
* \brief Merge the self-attention state with another state in place.
* \tparam DType The data type of v and v_other.
* \param v The partial v to be updated in-place. (n, h, d)
* \param s The logsumexp value to be updated in-place. (n, h)
* \param v_other The other v to be merged. (n, h, d)
* \param s_other The other logsumexp value to be merged. (n, h)
* \param seq_len The sequence length.
* \param num_heads The number of heads of v and v_other.
* \param head_dim The dimension of each head.
* \param mask Optional mask of whether to merge given sequences or not. (n)
* \param stream The CUDA stream to execute the kernel.
* \return status Indicates whether CUDA calls are successful
* \note Both s and s_other are logsumexp values with base 2.
*/
template <typename DType>
cudaError_t MergeStateInPlace(DType* v, float* s, DType* v_other, float* s_other, uint32_t seq_len,
uint32_t num_heads, uint32_t head_dim, uint8_t* mask = nullptr,
cudaStream_t stream = nullptr) {
DISPATCH_HEAD_DIM(head_dim, HEAD_DIM, {
constexpr uint32_t vec_size = std::max(16U / sizeof(DType), HEAD_DIM / 32U);
uint32_t bdx = HEAD_DIM / vec_size;
uint32_t bdy = num_heads;
dim3 nblks(seq_len);
dim3 nthrs(bdx, bdy);
auto kernel = MergeStateInPlaceKernel<vec_size, DType>;
void* args[] = {&v, &s, &v_other, &s_other, &mask, &num_heads, &head_dim};
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, 0, stream));
});
return cudaSuccess;
}
/*!
* \brief Merge self-attention states of a list of index sets.
* \tparam DTypeIn The data type of v.
* \tparam DTypeO The data type of v_merged.
* \param v The partial v of index sets. (n, num_index_sets, h, d)
* \param s The logsumexp value of index sets. (n, num_index_sets, h)
* \param v_merged The merged v of index sets union. (n, h, d)
* \param s_merged The merged logsumexp value of index sets union. (n, h)
* \param num_index_sets The number of index sets.
* \param seq_len The sequence length.
* \param num_heads The number of heads of v.
* \param head_dim The dimension of each head.
* \param stream The CUDA stream to execute the kernel.
* \return status Indicates whether CUDA calls are successful
* \note s are logsumexp values with base 2.
*/
template <typename DTypeIn, typename DTypeO>
cudaError_t MergeStates(DTypeIn* v, float* s, DTypeO* v_merged, float* s_merged,
uint32_t num_index_sets, uint32_t seq_len, uint32_t num_heads,
uint32_t head_dim, cudaStream_t stream = nullptr) {
DISPATCH_HEAD_DIM(head_dim, HEAD_DIM, {
constexpr uint32_t vec_size = std::max(16U / sizeof(DTypeIn), HEAD_DIM / 32U);
constexpr uint32_t bdx = HEAD_DIM / vec_size;
if (num_index_sets >= seq_len) {
constexpr uint32_t num_threads = 128;
constexpr uint32_t bdy = num_threads / bdx;
dim3 nblks(seq_len, num_heads);
dim3 nthrs(bdx, bdy);
constexpr uint32_t num_smem_stages = 4;
auto kernel =
MergeStatesLargeNumIndexSetsKernel<vec_size, bdx, bdy, num_smem_stages, DTypeIn, DTypeO>;
void* args[] = {&v, &s, &v_merged, &s_merged, &num_index_sets, &num_heads};
uint32_t smem_size =
num_smem_stages * bdy * head_dim * sizeof(DTypeIn) + num_threads * sizeof(float);
FLASHINFER_CUDA_CALL(
cudaFuncSetAttribute(kernel, cudaFuncAttributeMaxDynamicSharedMemorySize, smem_size));
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, smem_size, stream));
} else {
uint32_t bdy = num_heads;
dim3 nblks(seq_len);
dim3 nthrs(bdx, bdy);
auto kernel = MergeStatesKernel<vec_size, DTypeIn, DTypeO>;
void* args[] = {&v, &s, &v_merged, &s_merged, &num_index_sets, &num_heads, &head_dim};
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, 0, stream));
}
});
return cudaSuccess;
}
template <typename DTypeIn, typename DTypeO>
cudaError_t AttentionSum(DTypeIn* v, DTypeO* v_sum, uint32_t num_index_sets, uint32_t seq_len,
uint32_t num_heads, uint32_t head_dim, cudaStream_t stream = nullptr) {
DISPATCH_HEAD_DIM(head_dim, HEAD_DIM, {
constexpr uint32_t vec_size = std::max(16U / sizeof(DTypeIn), HEAD_DIM / 32U);
constexpr uint32_t bdx = HEAD_DIM / vec_size;
uint32_t bdy = num_heads;
dim3 nblks(seq_len);
dim3 nthrs(bdx, bdy);
auto kernel = AttentionSumKernel<vec_size, DTypeIn, DTypeO>;
void* args[] = {&v, &v_sum, &num_index_sets, &num_heads, &head_dim};
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, 0, stream));
});
return cudaSuccess;
}
template <typename DTypeIn, typename DTypeO, typename IdType>
cudaError_t VariableLengthMergeStates(DTypeIn* v, float* s, IdType* indptr, DTypeO* v_merged,
float* s_merged, uint32_t max_seq_len, uint32_t* seq_len,
uint32_t num_heads, uint32_t head_dim, bool enable_pdl,
cudaStream_t stream = nullptr) {
int dev_id = 0;
int num_sms = 0;
int num_blocks_per_sm = 0;
FLASHINFER_CUDA_CALL(cudaGetDevice(&dev_id));
FLASHINFER_CUDA_CALL(cudaDeviceGetAttribute(&num_sms, cudaDevAttrMultiProcessorCount, dev_id));
DISPATCH_HEAD_DIM(head_dim, HEAD_DIM, {
constexpr uint32_t vec_size = std::max(16U / sizeof(DTypeIn), HEAD_DIM / 32U);
constexpr uint32_t bdx = HEAD_DIM / vec_size;
constexpr uint32_t num_threads = 128;
constexpr uint32_t bdy = num_threads / bdx;
constexpr uint32_t num_smem_stages = 4;
uint32_t smem_size =
num_smem_stages * bdy * head_dim * sizeof(DTypeIn) + num_threads * sizeof(float);
auto kernel = PersistentVariableLengthMergeStatesKernel<vec_size, bdx, bdy, num_smem_stages,
DTypeIn, DTypeO, IdType>;
FLASHINFER_CUDA_CALL(cudaOccupancyMaxActiveBlocksPerMultiprocessor(&num_blocks_per_sm, kernel,
num_threads, smem_size));
num_blocks_per_sm = min(num_blocks_per_sm, ceil_div(max_seq_len * num_heads, num_sms));
dim3 nblks(num_sms * num_blocks_per_sm);
dim3 nthrs(bdx, bdy);
void* args[] = {&v, &s, &indptr, &v_merged, &s_merged, &max_seq_len, &seq_len, &num_heads};
FLASHINFER_CUDA_CALL(
cudaFuncSetAttribute(kernel, cudaFuncAttributeMaxDynamicSharedMemorySize, smem_size));
// PDL launch
if (enable_pdl) {
cudaLaunchAttribute attribute[1];
attribute[0].id = cudaLaunchAttributeProgrammaticStreamSerialization;
attribute[0].val.programmaticStreamSerializationAllowed = 1;
cudaLaunchConfig_t config;
config.attrs = attribute;
config.numAttrs = 1;
config.gridDim = nblks;
config.blockDim = nthrs;
config.dynamicSmemBytes = smem_size;
config.stream = stream;
FLASHINFER_CUDA_CALL(cudaLaunchKernelEx(&config, kernel, v, s, indptr, v_merged, s_merged,
max_seq_len, seq_len, num_heads));
} else {
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, smem_size, stream));
}
});
return cudaSuccess;
}
template <typename DTypeIn, typename DTypeO, typename IdType>
cudaError_t VariableLengthAttentionSum(DTypeIn* v, IdType* indptr, DTypeO* v_sum,
uint32_t max_seq_len, uint32_t* seq_len, uint32_t num_heads,
uint32_t head_dim, bool enable_pdl,
cudaStream_t stream = nullptr) {
int dev_id = 0;
int num_sms = 0;
int num_blocks_per_sm = 0;
FLASHINFER_CUDA_CALL(cudaGetDevice(&dev_id));
FLASHINFER_CUDA_CALL(cudaDeviceGetAttribute(&num_sms, cudaDevAttrMultiProcessorCount, dev_id));
DISPATCH_HEAD_DIM(head_dim, HEAD_DIM, {
constexpr uint32_t vec_size = std::max(16U / sizeof(DTypeIn), HEAD_DIM / 32U);
constexpr uint32_t bdx = HEAD_DIM / vec_size;
constexpr uint32_t num_threads = 128;
constexpr uint32_t bdy = num_threads / bdx;
constexpr uint32_t num_smem_stages = 4;
uint32_t smem_size = num_smem_stages * bdy * head_dim * sizeof(DTypeIn);
auto kernel = PersistentVariableLengthAttentionSumKernel<vec_size, bdx, bdy, num_smem_stages,
DTypeIn, DTypeO, IdType>;
FLASHINFER_CUDA_CALL(cudaOccupancyMaxActiveBlocksPerMultiprocessor(&num_blocks_per_sm, kernel,
num_threads, smem_size));
num_blocks_per_sm = min(num_blocks_per_sm, ceil_div(max_seq_len * num_heads, num_sms));
dim3 nblks(num_sms * num_blocks_per_sm);
dim3 nthrs(bdx, bdy);
void* args[] = {&v, &indptr, &v_sum, &max_seq_len, &seq_len, &num_heads};
FLASHINFER_CUDA_CALL(
cudaFuncSetAttribute(kernel, cudaFuncAttributeMaxDynamicSharedMemorySize, smem_size));
if (enable_pdl) {
// PDL launch
cudaLaunchAttribute attribute[1];
attribute[0].id = cudaLaunchAttributeProgrammaticStreamSerialization;
attribute[0].val.programmaticStreamSerializationAllowed = 1;
cudaLaunchConfig_t config;
config.attrs = attribute;
config.numAttrs = 1;
config.gridDim = nblks;
config.blockDim = nthrs;
config.dynamicSmemBytes = smem_size;
config.stream = stream;
FLASHINFER_CUDA_CALL(
cudaLaunchKernelEx(&config, kernel, v, indptr, v_sum, max_seq_len, seq_len, num_heads));
} else {
FLASHINFER_CUDA_CALL(cudaLaunchKernel((void*)kernel, nblks, nthrs, args, smem_size, stream));
}
});
return cudaSuccess;
}
} // namespace flashinfer
#endif // FLASHINFER_CASCADE_CUH_
|